#include	<dos.h>
#include	<conio.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<bios.h>

#define	YES	1		// Useful booleans
#define NO	0
#define ON	1
#define OFF	0
#define ESC	0x1b		// ASCII code for escape key
#define ACK	0xFA		// Keyboard acknowledge at port 0x60
#define RESEND	0xFE		// Keyboard resend request

unsigned char * convert_to_binary(unsigned char number);
unsigned char setbits(unsigned char a, int n, int k, char v);
unsigned char getbit(unsigned char a, int n);
unsigned char setbit(unsigned char a, int n, int which);

void interrupt far (*old_09)(void);

void	reset_keyboard(void);
void 	interrupt far new_09(void);
void	test_1(void);
void	test_2(void);
void	test_3(void);
void	num_lock(void);
void	caps_lock(void);
void	scrollock(void);
void	command_byte(unsigned char cmd);
void	send_command_byte(void);
void	done(void);
void	introduction(void);
void	next_half(void);
void	test_2a(void);
void	test_3a(void);
void	do_it(void);


unsigned char far *status_byte = NULL;
unsigned char far *shift_flag = NULL;
unsigned char far *extended_shift_flag = NULL;
unsigned char response;

void	main(void)

{

introduction();		// Put up intro screen
disable();		// No interrupts
old_09 = getvect(0x09);	// Save old keyboard interrupt
setvect(0x09, new_09);	// New ISR
enable();		// Re-enable interrupts
test_1();		// Print status bytes and reset keyboard
test_2();		// Cycle the LEDs with the 8042 controller
reset_keyboard();	// Reset the keyboard
test_3();		// Set typematic/repeat rates with 8042 controller
reset_keyboard();       // Reset the keyboard
disable();
setvect(0x09,old_09);	// Original ISR
enable();
next_half();		// Intro screen for next half
test_2a();		// Cycle LEDs with BIOS calls
test_3a();		// Set typematic/repeat rates with BIOS
reset_keyboard();	// Back to default settings
done();


}

void	introduction(void)


{
clrscr();
printf("This keyboard diagnostic program key_8042 is FREEWARE, written");
printf("\nby Walt Myers, COMPUSERVE ID 76407,2235.");
printf("\n\n\nThe 1st half of this program checks your keyboard \
functions by programming the\n");
printf("8042 controller and by taking over the int 0x09 keyboard ISR.");
printf("\nNo checks are made using DOS or BIOS calls.  All checks are\
 done at");
printf("\n8042 hardware level. ");
printf("\n\nThe 2nd half does the same checks but only uses BIOS calls.");
printf("\n\nPress any key to continue...");
getch();


}


void 	interrupt far new_09(void)

{

response = inportb(0x60);	// Get keyboard response first
(*old_09)();			// Chain to old interrupt

}



void	test_1(void)

{

char result = 0;
unsigned char *converted = NULL;

clrscr();

status_byte = MK_FP(0x0040,0x0097);		// LED status byte location
shift_flag  = MK_FP(0x0040,0x0017);		// Shift flag location
extended_shift_flag  = MK_FP(0x0040,0x0018);	// EXT shift flag location

converted = convert_to_binary(*status_byte);	// Write as binary
printf("LED status byte:         %s\n",converted);


printf("\t7  Error Flag for keyboard command  6  LED update in progress\n");
printf("\t5  RESEND received from keyboard    4  ACK received from keyboard\n");
printf("\t3  Reserved                         2  CAPSLOCK LED status\n");
printf("\t1  NUMLOCK LED status               0  SCROLLOCK LED status\n\n");



converted = convert_to_binary(*shift_flag);
printf("shift flag:          %s\n", converted);

printf("\t7  = 1 Insert active                6  = 1 Caps lock active\n");
printf("\t5  = 1 Num lock active              4  = 1 Scroll lock active\n");
printf("\t3  = 1 ALT pressed                  2  = 1 CTRL pressed\n");
printf("\t1  = 1 Left Shift pressed           0  = 1 Right Shift pressed\n\n");


converted = convert_to_binary(*extended_shift_flag);
printf("extended shift flag: %s\n",converted);

printf("\t7  = 1 Insert pressed               6  = 1 Caps Lock pressed\n");
printf("\t5  = 1 Num Lock pressed             4  = 1 Scroll Lock pressed\n");
printf("\t3  = 1 CTRL-NUMLOCK state active    2  = 1 SYS REQ pressed\n");
printf("\t1  = 1 Left ALT pressed             0  = 1 Left CTRL pressed\n\n");




result = inportb(0x64);				// Get status register
converted = convert_to_binary(result);
printf("Status Register:     %s\n",converted);
printf("\nAbout to issue reset:  Watch for 3 LEDs. Press any key.");
getch();

reset_keyboard();

}

void	test_2(void)

{
int i;

clrscr();
printf("This next test will cycle the NUM, CAPS & SCROLL LOCK keys\
 using the 8042\nregisters only.");
printf("\n\nThere'll be 20 SLOW cycles.  Press any key to start...\n\
Cycle #  0 /20 total");
getch();

for (i = 0; i < 20; i++)

	{
		send_command_byte();	//	Send 0xed
		num_lock();		//	Send data byte
		send_command_byte();
		caps_lock();
		send_command_byte();
		scrollock();
		gotoxy(10,5);
		printf("%d",i+1);
	}


}



void	send_command_byte(void)

{

response = RESEND;      			// Init flag

while (response == RESEND) command_byte(0xED);	// Send Blink LED cmd

while (response != ACK);			// Wait for acknowledge

}


void	test_3(void)

{

char key = 0;

clrscr();
printf("This test will set typematic/repeat rates to FAST/FAST, let you");
printf("\ntest on the keyboard, then SLOW/SLOW and let you test.");
printf("\nPress any key to start....");

//		First, we set the keyboard to go fast

response = RESEND;				// Init flag

while (response == RESEND) command_byte(0xF3);	// Send command byte

while (response != ACK);

outportb(0x60,(char)0);		//	Send go FAST data byte

printf("\n\nKeyboard is FAST.  Test typematic/delay rate.  \
Press ESC to continue...\n");

while(key != ESC)		//	Allow user keyboard input

{
_AH = 0x00;
geninterrupt(0x16);		//	Get keyboard input
key = _AL;			//	Save register
putchar((int) key);		//	Display character to screen

}

//		Next, set the keyboard to go real slow

response = RESEND;				// Init flag

while (response == RESEND) command_byte(0xF3);	// Send command byte

while (response != ACK);

outportb(0x60,(char)127);			// Send go SLOW data byte

printf("\n\nKeyboard is SLOW.  Test typematic/delay rate.  \
Press ESC to continue...\n\n");

key = 0;

while(key != ESC)

{
_AH = 0x00;
geninterrupt(0x16);		//	Get keyboard input
key = _AL;			//	Save register
putchar((int) key);		//	Display character to screen

}


}

unsigned char * convert_to_binary(unsigned char number)

{

static unsigned char buffer[9];
unsigned int mask = 0x80;
int x;

for (x = 0; x < 8; x++) buffer[x] = (number & (mask >> x)) ? '1' : '0';
buffer[8] = 0;
return(buffer);

}

unsigned char setbits(unsigned char a, int n, int k, char v)

{
return(a & ~(~(~(unsigned char)0 << k) << n)) | (v << n);

}

unsigned char getbit(unsigned char a, int n)

{

return(a >> n) & 01;

}

unsigned char setbit(unsigned char a, int n, int which)

{

if (which == ON) return(a | (01 << n));
return(a & ~(01 << n));

}


void	num_lock(void)

{
unsigned long int x=0L;


outportb(0x60,0x02);		//	Send data byte

for (x = 0L; x < 500000L; x++);	//	Wait an AWFUL long time

}


void	caps_lock(void)

{

unsigned long int x=0L;

outportb(0x60,0x04);

for (x = 0L; x < 500000L; x++);

}

void	scrollock(void)

{

unsigned long int x=0L;

outportb(0x60,0x01);

for (x = 0L; x < 500000L; x++);

}

void	command_byte(unsigned char cmd)

{

outportb(0x60,cmd);

}

void	reset_keyboard(void)

{

outportb(0x60,0xFF);				// Reset keyboard

}


void	done(void)

{

clrscr();

printf("\nProgram terminated.\n\n");
printf("This keyboard diagnostic program key_8042 is FREEWARE, written");
printf("\nby Walt Myers, COMPUSERVE ID 76407,2235");


}

void	next_half(void)

{
clrscr();

printf("For the 2nd half, do as many of the checks as possible by using\
 only BIOS calls.");
printf("\n\nKeyboard vector now restored.  Press any key to start....");
getch();

}


void	test_2a(void)		//	Blink the LEDs

{

int i;

clrscr();
printf("This next test will cycle the NUM, CAPS & SCROLL LOCK keys\
 using BIOS only.");
printf("\n\nThere'll be 20 SLOW cycles.  Press any key to start...\n\
Cycle #  0 /20 total");
getch();

*shift_flag = setbits(*shift_flag,4,3,0);	// Set all to OFF
do_it();					// Update LEDs


for (i = 0; i < 20; i ++)	//	Blink 20 times

{
	*shift_flag = setbit(*shift_flag,5,ON);	// Turn on NUMLOCK
	do_it();
	*shift_flag = setbit(*shift_flag,5,OFF);// Turn it off
	do_it();
	*shift_flag = setbit(*shift_flag,6,ON);	// Turn on CAPS
	do_it();
	*shift_flag = setbit(*shift_flag,6,OFF);// Turn it off
	do_it();
	*shift_flag = setbit(*shift_flag,4,ON);	// Turn on SCROLLOCK
	do_it();
	*shift_flag = setbit(*shift_flag,4,OFF);// Turn it off
	do_it();
	gotoxy(10,4);
	printf("%d",i+1);

}

*shift_flag = setbits(*shift_flag,4,3,0);	// Set all to OFF
do_it();					// Update LEDs


}

void	do_it(void)

{

unsigned long int x;

_AH = 0x01;        		// Return keyboard status/update LEDs
geninterrupt(0x16);		// Keyboard DSR

for (x = 0L; x < 300000L; x++);	// Wait a long time

}


void	test_3a(void)		//	Set typematic/repeat rates

{

char key = 0;

clrscr();
printf("This test will set typematic/repeat rates to FAST/FAST, let you");
printf("\ntest on the keyboard, then SLOW/SLOW and let you test.");
printf("\nPress any key to start....");

//		First, we set the keyboard to go fast

_AX = 0x0305;			//	Set typematic/repeat rates
_BX = 0x0000;			//	Go real fast
geninterrupt(0x16);		//	Make it happen


printf("\n\nKeyboard is FAST.  Test typematic/delay rate.  \
Press ESC to continue...\n");

while(key != ESC)		//	Allow user keyboard input

{
_AH = 0x00;
geninterrupt(0x16);		//	Get keyboard input
key = _AL;			//	Save register
putchar((int) key);		//	Display character to screen

}

//		Next, set the keyboard to go real slow

_AX = 0x0305;			//	Set typematic/repeat rate
_BX = 0x031F;			//	Go real slow
geninterrupt(0x16);


printf("\n\nKeyboard is SLOW.  Test typematic/delay rate.  \
Press ESC to continue...\n\n");

key = 0;

while(key != ESC)

{
_AH = 0x00;
geninterrupt(0x16);		//	Get keyboard input
key = _AL;			//	Save register
putchar((int) key);		//	Display character to screen

}

}

